"use strict";

var $ = require("../../core/renderer"),
    modules = require("./ui.grid_core.modules"),
    extend = require("../../core/utils/extend").extend,
    FilterBuilder = require("./../filter_builder"),
    messageLocalization = require("../../localization/message"),
    ScrollView = require("./../scroll_view"),
    Popup = require("./../popup");

var FilterBuilderView = modules.View.inherit({
    _renderCore: function _renderCore() {
        this._updatePopupOptions();
    },

    _updatePopupOptions: function _updatePopupOptions() {
        if (this.option("filterBuilderPopup.visible") && !this._filterBuilderPopup) {
            this._initPopup();
        } else {
            this._filterBuilderPopup && this._filterBuilderPopup.option(this._getPopupOptions());
        }
    },

    _getPopupOptions: function _getPopupOptions() {
        var that = this;

        return extend({
            title: messageLocalization.format("dxDataGrid-filterBuilderPopupTitle"),
            contentTemplate: function contentTemplate($contentElement) {
                return that._getPopupContentTemplate($contentElement);
            },
            onShowing: function onShowing() {
                if (that._filterBuilder.option("value") !== that.option("filterValue")) {
                    that._filterBuilder.option("value", that.option("filterValue"));
                }
            },
            onOptionChanged: function onOptionChanged(args) {
                if (args.name === "visible") {
                    that.option("filterBuilderPopup.visible", args.value);
                }
            },
            toolbarItems: that._getPopupToolbarItems()
        }, that.option("filterBuilderPopup"));
    },

    _initPopup: function _initPopup() {
        var that = this,
            $popupContainer = this.element();
        that._filterBuilderPopup = that._createComponent($popupContainer, Popup, extend(that._getPopupOptions()));
    },

    _getPopupContentTemplate: function _getPopupContentTemplate($contentElement) {
        var $filterBuilderContainer = $("<div>").appendTo($contentElement),
            fields = this.getController("columns").getColumns(),
            customOperations = this.getController("filterSync").getCustomFilterOperations();

        fields = fields.filter(function (item) {
            return item.allowFiltering;
        }).map(function (item) {
            var column = extend({}, item, { filterOperations: null });
            return column;
        });

        this._filterBuilder = this._createComponent($filterBuilderContainer, FilterBuilder, extend(this.option("filterBuilder"), {
            customOperations: customOperations,
            fields: fields
        }));

        this._createComponent($contentElement, ScrollView, { direction: "both" });
    },

    _getPopupToolbarItems: function _getPopupToolbarItems() {
        var that = this;
        return [{
            toolbar: "bottom",
            location: "after",
            widget: "dxButton",
            options: {
                text: messageLocalization.format("OK"),
                onClick: function onClick(e) {
                    var filter = that._filterBuilder.option("value");
                    that.option("filterValue", filter);
                    that._filterBuilderPopup.hide();
                }
            }
        }, {
            toolbar: "bottom",
            location: "after",
            widget: "dxButton",
            options: {
                text: messageLocalization.format("Cancel"),
                onClick: function onClick(e) {
                    that._filterBuilderPopup.hide();
                }
            }
        }];
    },

    optionChanged: function optionChanged(args) {
        switch (args.name) {
            case "filterBuilder":
            case "filterBuilderPopup":
                this._invalidate();
                args.handled = true;
                break;
            default:
                this.callBase(args);
        }
    }
});

module.exports = {
    defaultOptions: function defaultOptions() {
        return {
            /**
             * @name GridBaseOptions_filterBuilder
             * @publicName filterBuilder
             * @type dxFilterBuilderOptions
             * @default {}
             */
            filterBuilder: {},

            /**
             * @name GridBaseOptions_filterPopup
             * @publicName filterPopup
             * @type dxPopupOptions
             * @default {}
             */
            filterBuilderPopup: {}
        };
    },
    views: {
        filterBuilderView: FilterBuilderView
    }
};